#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>
#include "jewels.h"

static int map[8][8];
static short int next[8][1025];
static int next_i[8] = {0};

static time_t end_time = 0;

struct cursor {
	int x;
	int y;
	int h;
} cur;

void load_input(char *filename, int (*map)[8][8])
{
	FILE *file = fopen(filename, "r");
	if (file == NULL) {
		printf("Soubor %s nelze otevrit!\n", filename);
		exit(1);
	}

	for (int i = 0; i < 8; i++)
		for (int j = 0; j < 8; j++)
			fscanf(file, "%1d ", &(*map)[i][j]);

	/* Dalsi kameny */
	for (int i = 0; i < 8; i++) {
		int pos = 0;
		while (1) {
			char c = getc(file);
			if (c == '\r')
				continue;
			if (c == '\n' || c == EOF)
				break;
			putchar(c);
			next[i][pos++] = c - '0';
		}
			putchar('\n');
	}

	fclose(file);
}

void draw_segment(int key, int x, int y)
{
	attron(COLOR_PAIR(key));
	mvaddstr(x, y, "    ");
	mvaddstr(x+1, y, "    ");
	attroff(COLOR_PAIR(key));
}

void draw_map(void)
{
	const int x = 1, y = 8;

	for (int i = 0; i < 8; i++)
		for (int j = 0; j < 8; j++)
			draw_segment(map[i][j], y+i*2, x+j*4);
}

void draw_help(void)
{
	const int x = 35, y = 8;

	mvprintw(y+0, x, "    Q ... ukoncit program");
	mvprintw(y+1, x, "sipky ... pohyb kurzoru");
	mvprintw(y+2, x, "    S ... zmena orientace kurzoru");
	mvprintw(y+3, x, "    D ... prohozeni kamenu");
}

time_t parse_time(char *str)
{
	int min, sec;
	sscanf(str, "%d:%d", &min, &sec);
	return min*60+sec;
}

void game_over(void)
{
	mvaddstr(14,  11, "           ");
	mvaddstr(15,  11, " KONEC HRY ");
	mvaddstr(16, 11, "           ");
	while (tolower(getch()) != 'q')
		usleep(20000);
	endwin();
	exit(0);
}

void draw_time(void)
{
	if (end_time) {
		time_t diff = end_time - time(NULL);
		if (diff >= 0)
			mvprintw(1, 8, "Zbyvajici cas %d:%d   ", (int)(diff/60), diff%60);
		if (diff <= 0)
			game_over();
	}
}

void draw_cursor(void)
{
	const int x = 1, y = 8;

	attron(COLOR_PAIR(map[cur.x][cur.y]));
	attron(A_REVERSE);

	mvaddch(y+cur.y*2, x+cur.x*4, '/');
	mvaddch(y+(cur.y+!cur.h)*2+1, x+cur.x*4, '\\');
	mvaddch(y+(cur.y+!cur.h)*2+1, x+(cur.x+cur.h)*4+3, '/');
	mvaddch(y+cur.y*2, x+(cur.x+cur.h)*4+3, '\\');

	attroff(COLOR_PAIR(map[cur.x][cur.y]));
	attroff(A_REVERSE);
}

void remove_seqs(int (*map)[8][8])
{
	char rm[8][8] = {0};

	/* Horizontalni */
	for (int i = 0; i < 8; i++)
		for (int j = 0; j < 5; j++)
			if ((*map)[i][j] == (*map)[i][j+1] && (*map)[i][j] == (*map)[i][j+2])
				rm[i][j] = rm[i][j+1] = rm[i][j+2] = 1;

	/* Vertikalni */
	for (int i = 0; i < 5; i++)
		for (int j = 0; j < 8; j++)
			if ((*map)[i][j] == (*map)[i+1][j] && (*map)[i][j] == (*map)[i+2][j])
				rm[i][j] = rm[i+1][j] = rm[i+2][j] = 1;

	/* Gravitace */
	for (int j = 0; j < 8; j++) {
		for (int i = 6; i >= 0; i--) {
			if (!rm[i][j]) {
				int row = i;
				while (rm[row+1][j] && row < 6) {
					swap(map, row, j, row+1, j);
					rm[row+1][j] = 0;
					rm[row][j] = 1;
					row++;
				}
			}
		}
	}

	/* Doplneni */
	for (int i = 0; i < 8; i++)
		for (int j = 8; j >= 0; j--)
			if (rm[i][j])
				(*map)[i][j] = next[i][next_i[i]++];

}

void draw_next(void)
{
	for (int i = 0; i < 8; i++)
		draw_segment(next[i][next_i[i]], 5, 1+i*4);
	mvaddstr(7, 1, "================================");
}

void play(void)
{
	initscr();
	raw();
	keypad(stdscr, TRUE);
	noecho();
	curs_set(0);
	nodelay(stdscr, TRUE);
	start_color();

	init_pair(1, COLOR_BLACK, COLOR_RED);
	init_pair(2, COLOR_BLACK, COLOR_GREEN);
	init_pair(3, COLOR_BLACK, COLOR_BLUE);
	init_pair(4, COLOR_BLACK, COLOR_WHITE);
	init_pair(5, COLOR_BLACK, COLOR_YELLOW);
	init_pair(6, COLOR_BLACK, COLOR_MAGENTA);

	cur = (struct cursor){0, 0, 1};

	/* Zacni pocitat cas */
	if (end_time)
		end_time += time(NULL);

	draw_map();
	draw_cursor();
	draw_help();
	draw_time();
	//draw_next();
	refresh();

	while (1) {
		int c = getch();
		switch (c) {
		case 'q':
		case 'Q':
			/* Konec */
			endwin();
			return;
		case 's':
		case 'S':
			/* Zmen orientaci kurzoru */
			if ((cur.h && cur.y < 7)
			|| (!cur.h && cur.x < 7))
				cur.h = !cur.h;
			else
				continue;
			break;
		case 'd':
		case 'D':
			/* Prohoz kamenu */
			if (valid_move(&map, cur.y, cur.x, cur.y+!cur.h, cur.x+cur.h))
				swap(&map, cur.y, cur.x, cur.y+!cur.h, cur.x+cur.h);
			draw_map();
			refresh();
			while (has_seq(&map)) {
				usleep(500000);
				remove_seqs(&map);
			}
			break;
		case KEY_RIGHT:
			if (cur.x < 7)
				cur.x++;
			break;
		case KEY_LEFT:
			if (cur.x > 0)
				cur.x--;
			break;
		case KEY_DOWN:
			if (cur.y < 7)
				cur.y++;
			break;
		case KEY_UP:
			if (cur.y > 0)
				cur.y--;
			break;
		default:
			draw_time();
			usleep(10000);
			continue;
		}
		draw_map();
		//draw_next();
		draw_cursor();
		/*if (valid_move(&map, cur.y, cur.x, cur.y+!cur.h, cur.x+cur.h))
			mvaddstr(10, 35, "  lze tahnout");
		else
			mvaddstr(10, 35, "nelze tahnout");*/
		refresh();
	}
}

int main(int argc, char **argv)
{
	if (argc < 2 || argc > 3) {
		printf("Pouziti: %s mapovy_soubor [cas]\n", argv[0]);
		return 1;
	}
	if (argc == 3)
		end_time = parse_time(argv[2]);
	load_input(argv[1], &map);
	//print_map(&map);
	play();
	return 0;
}